#include <stdio.h>
#include "sock.h"
#include "parser.h"
#include "display.h"
#include "main.h"

HANDLE						hEventStop;					// Handle used to handle Ctrl+C/Break event
DWORD						dwTimeOut = INFINITE;		// Time-Out used for the ping/retransmit mechanism
int							number_phone_table = 0;		// Number of registred phones
struct phone_table_entry	*phone_table;				// List of registred phones

// Ctrl+C/Break event handler
void SigHandler()
{
	puts("Catch SIG_TERM. Calling cleanup fonction");
	exit(0);
}

int main(void)
{
	HANDLE	hListEvent[2];
	u_short	iPort = 5000;

	puts_3("i2004 Server");

	// Setting up the Ctrl+C/Break handler
	signal(SIG_TERM, SigHandler);

	// Starting the network handler
	if(init_sock(iPort) < 0)
	{
		puts_1("Aborting server startup");
		shutdown_sock();
		return;
	}

	// Allocate memory for the phone list
	phone_table = malloc(sizeof(struct phone_table_entry) * MAX_PHONE);
	if(!phone_table)
	{
		puts_1("Not enough memory (phone_table)");
		return;
	}

	// Ready - waiting for an event
	printf_3("Listening on UDP Port %d\n", iPort);
	hListEvent[0] = hEventStop;
	hListEvent[1] = wsaOverlappedRecv.hEvent;
	do
	{
		DWORD tick;
		int i;

		// Looking for the smallest time-out value
		tick = GetTickCount();
		dwTimeOut = INFINITE;
		for(i = 0 ; i < number_phone_table ; i++)
		{
			if(dwTimeOut > phone_table[i].timeout - tick)
				dwTimeOut = phone_table[i].timeout - tick;
		}
		switch(WaitForMultipleObjects(2, hListEvent, FALSE, dwTimeOut))
		{
			case WAIT_FAILED :	DisplayLastError("Error wainting events");
			case WAIT_OBJECT_0: // This is the stop event
								puts_3("Shutdown...");
								shutdown_sock();
								for(i = 0 ; i < number_phone_table ; i++)
								{
									free(phone_table[i].ptr_phone_entry);
								}
								free(phone_table);
								CloseHandle(hEventStop);
								return;
			case WAIT_OBJECT_0 + 1: // A data is ready in the incoming buffer
								read_data();
								break;
			case WAIT_TIMEOUT:  // looking for the reason of the timeout
								puts_4("Time-Out in WaitForMultipleObjects");
								tick = GetTickCount();
								for(i = 0 ; i < number_phone_table ; i++)
								{
									if(phone_table[i].timeout <= tick)
									{
										// If the queue is empty, send a ping
										if(phone_table[i].ptr_phone_entry->last_buf_available == 0)
											SendPing(&phone_table[i]);
										else
											SendRetransmit(&phone_table[i]);
									}
								} // for
								break;
		} // switch WaitForMultipleObjects
	}while(1);
	return;
}
