#include <winsock2.h>
#include <windows.h>
#include "display.h"
#include "sock.h"
#include "parser.h"


SOCKET				hostSocket = 0;						// Socket handle used for everything (send & receive)
struct sockaddr_in	addr_from;							// Global : give the IP address for the last packet received
int					size_addr_from = sizeof(addr_from);	// Give the size of the sockaddr_in (used by WSARecvFrom)
char				*buff;								// Receive buffer address

extern struct phone_table_entry	*phone_table;
extern int						number_phone_table;
extern DWORD					dwTimeOut;

int init_sock(u_short iPort)
{
	WSADATA			WSAData;
	struct sockaddr_in address;
	DWORD flag = 0;
	DWORD tmp;

	puts_3("Starting the socket layer...");
	// Starting Winsock
	if (WSAStartup(MAKEWORD(2,0), &WSAData) != 0)
	{
		DisplayLastError("Error while initializing the socket layer");
		return -1;
	}

	puts_3("Creating server socket...");
	if((hostSocket = WSASocket(AF_INET, SOCK_DGRAM, 0, NULL, 0, WSA_FLAG_OVERLAPPED)) == INVALID_SOCKET)
		DisplayLastErrorWSA("Error creating socket");

	memset(&address, 0, sizeof(address));
	address.sin_family = AF_INET;
	address.sin_addr.s_addr = INADDR_ANY;
	address.sin_port = htons(iPort);

	puts_3("Binding the server socket...");
	if (bind(hostSocket, (struct sockaddr *) &address, sizeof(address)) < 0)
	{
		DisplayLastErrorWSA("Error during the bind() call");
		return -2;
	}
	if((buff = VirtualAlloc(NULL, SIZE_PAGE, MEM_COMMIT, PAGE_READWRITE)) == 0)
	{
		DisplayLastError("Error allocating network buffer");
		return -3;
	}
	if((wsaOverlappedRecv.hEvent = WSACreateEvent()) == WSA_INVALID_EVENT)
	{
		DisplayLastErrorWSA("Error creating an event");
		return -4;
	}
	wsabuf.buf = buff;
	wsabuf.len = SIZE_PAGE;
	if(WSARecvFrom(hostSocket, &wsabuf, 1, &tmp, &flag, (struct sockaddr *)&addr_from, &size_addr_from, &wsaOverlappedRecv, NULL) == SOCKET_ERROR)
		if(WSAGetLastError() != WSA_IO_PENDING)
			DisplayLastErrorWSA("Error WSARecvFrom");
	return 0;
}

void shutdown_sock()
{
	if(hostSocket)
		closesocket(hostSocket);
	if(wsaOverlappedRecv.hEvent)
		CloseHandle(wsaOverlappedRecv.hEvent);
	WSACleanup();
	return;
}

void read_data()
{
	DWORD dwNumberOfBytesRecv, dwFlagsRecv;
	DWORD dwNumberOfBytesRecvd, flag = 0;
	int   tmp;

	WSAResetEvent(wsaOverlappedRecv.hEvent);
	if(WSAGetOverlappedResult(hostSocket, &wsaOverlappedRecv, &dwNumberOfBytesRecv, FALSE, &dwFlagsRecv) == FALSE)
		DisplayLastErrorWSA("Error WSAGetOverlappedResult");
	
	// Looking in the phone list if we already have a registration for him
	for(tmp = 0 ; tmp < number_phone_table ; tmp ++)
	{
		if(phone_table[tmp].addr_from.sin_addr.S_un.S_addr == addr_from.sin_addr.S_un.S_addr)
			break;
	}
	// DEBUG CODE : dump packet
	printf_4("\n*** Dump %d bytes from %s - phone_table[%d] ***\n", dwNumberOfBytesRecv, inet_ntoa(addr_from.sin_addr), tmp);
#ifdef DUMP_PACKET
	for(dwNumberOfBytesRecvd=0;dwNumberOfBytesRecvd<dwNumberOfBytesRecv; dwNumberOfBytesRecvd++)
		printf("%.2x ", (unsigned char)wsabuf.buf[dwNumberOfBytesRecvd]);
	puts("\n******************************************");
#endif

	if(tmp >= number_phone_table)
	{
		puts_3("Received a packet from an unknown source");
		parsing(dwNumberOfBytesRecv, wsabuf.buf, NULL, (struct sockaddr_in *)&addr_from);
	}
	else
		parsing(dwNumberOfBytesRecv, wsabuf.buf, &phone_table[tmp], (struct sockaddr_in *)&addr_from);
	if(WSARecvFrom(hostSocket, &wsabuf, 1, &dwNumberOfBytesRecvd, &flag, (struct sockaddr *)&addr_from, &size_addr_from, &wsaOverlappedRecv, NULL) == SOCKET_ERROR)
		if(WSAGetLastError() != WSA_IO_PENDING)
			DisplayLastErrorWSA("Error WSARecvFrom");
}

// Allocate memory & initialize structures for a new phone
int CreateClient(struct sockaddr_in *addr_from)
{
	int tmp;
	if(number_phone_table >= MAX_PHONE)
	{
		puts_1("Error : can't handle a new phone : too many phones. Increase the MAX_PHONE value");
		return 0;
	}
	phone_table[number_phone_table].ptr_phone_entry = malloc(sizeof(struct phone_entry));
	if(!phone_table[number_phone_table].ptr_phone_entry)
	{
		puts_1("Error : malloc phone_entry");
		return 0;
	}
	phone_table[number_phone_table].timeout = 0;
	memcpy(&(phone_table[number_phone_table].addr_from), addr_from, sizeof(struct sockaddr_in));
	phone_table[number_phone_table].ptr_phone_entry->seq_phone = (short)0x8485;
	phone_table[number_phone_table].ptr_phone_entry->seq_server = (short)0xbc20;
	phone_table[number_phone_table].ptr_phone_entry->last_seq_ack = (short)0xbc20;
	phone_table[number_phone_table].ptr_phone_entry->last_buf_available = 0;
	phone_table[number_phone_table].ptr_phone_entry->nb_retransmit = 0;
	phone_table[number_phone_table].ptr_phone_entry->state = STATE_INIT;
	phone_table[number_phone_table].ptr_phone_entry->receiver_state = STATE_HANGUP;
	phone_table[number_phone_table].ptr_phone_entry->tick_next_ping = GetTickCount() + PING_TIMER;
	// Initialize WSABUF
	for(tmp = 0; tmp < MAX_BUF_NUMBER ; tmp++)
	{
		phone_table[number_phone_table].ptr_phone_entry->wsabufsend[tmp].buf = phone_table[number_phone_table].ptr_phone_entry->buf[tmp];
	}
	return number_phone_table++;
}

// Send data to a phone without retransmit nor buffering
void SendRawClient(int size, char *data, struct sockaddr_in *addr_from)
{
	DWORD tmp;
	WSABUF wsabufsend;

	wsabufsend.buf = data;
	wsabufsend.len = size;

	if(WSASendTo(hostSocket, &wsabufsend, 1, &tmp, 0, (struct sockaddr *)addr_from, sizeof(*addr_from),
						NULL, NULL) == SOCKET_ERROR)
		DisplayLastErrorWSA("Error sending datas");
}

// Send data to a phone with retransmit+buffering
void SendClient(int size, char *data, struct phone_table_entry *pte)
{
	DWORD tmp, tick;
	int	  buf_pos = pte->ptr_phone_entry->last_buf_available, i;
	unsigned short *sdata = (unsigned short*)data;

	if(buf_pos >= MAX_BUF_NUMBER)
	{
		puts_1("Error : send queue overflow");
		return;
	}
	sdata[1] = htons(++(pte->ptr_phone_entry->seq_server));
	pte->ptr_phone_entry->wsabufsend[buf_pos].len = size;
	memcpy(pte->ptr_phone_entry->wsabufsend[buf_pos].buf, data, size);
	pte->ptr_phone_entry->wsaoverlap[buf_pos].hEvent = 0;

	tick = GetTickCount();
	pte->timeout = tick + RETRANSMIT_TIMER;
	dwTimeOut = INFINITE;
	for(i = 0 ; i < number_phone_table ; i++)
	{
		if(dwTimeOut > phone_table[i].timeout - tick)
			dwTimeOut = phone_table[i].timeout - tick;
	}

	// DEBUG CODE : dump packet
	printf_4("\n**> Send %d bytes to %s using slot #%d ***\n", pte->ptr_phone_entry->wsabufsend[buf_pos].len, inet_ntoa(pte->addr_from.sin_addr), buf_pos);
#ifdef DUMP_PACKET
	for(tmp=0;tmp<pte->ptr_phone_entry->wsabufsend[buf_pos].len; tmp++)
		printf("%.2x ", (unsigned char)pte->ptr_phone_entry->wsabufsend[buf_pos].buf[tmp]);
	puts("\n******************************************");
#endif


	if(WSASendTo(hostSocket, &(pte->ptr_phone_entry->wsabufsend[buf_pos]), 1, &tmp, 0, (struct sockaddr *)&(pte->addr_from), sizeof(struct sockaddr_in),
						&(pte->ptr_phone_entry->wsaoverlap[buf_pos]), NULL) == SOCKET_ERROR)
		DisplayLastErrorWSA("Error sending datas");
	pte->ptr_phone_entry->last_buf_available++;
}


// Free memory allocated for a phone
void CloseClient(struct phone_table_entry *pte)
{
	int i,j,k;

	// Notify other phones if we're in their favorite list
	for(j = 0 ; j < number_phone_table ; j++)
	{
		for(k = 1 ; k <= 5 ; k++)
		{
			if(memcmp(pte->ptr_phone_entry->mac_current, phone_table[j].ptr_phone_entry->mac_fav[k], 17) == 0)
			{
				SendFavorite((unsigned char)k, 0x00, &phone_table[j], phone_table[j].ptr_phone_entry->name_fav[k]);
				break;
			}
		}
	}

	free(pte->ptr_phone_entry);
	for(i = 0 ; i < number_phone_table ; i++)
	{
		if(phone_table[i].addr_from.sin_addr.S_un.S_addr == pte->addr_from.sin_addr.S_un.S_addr)
			break;
	}
	if(i >= number_phone_table)
	{
		puts_2("Error : freeing an unknown client");
		return;
	}
	number_phone_table--;
	printf_3("Client #%d freed\n", i);
	if(i == number_phone_table)
		return;
	phone_table[i].addr_from = phone_table[number_phone_table].addr_from;
	phone_table[i].ptr_phone_entry = phone_table[number_phone_table].ptr_phone_entry;
	phone_table[i].timeout = phone_table[number_phone_table].timeout;
	return;
}

